"""Interceptor that collects metadata from incoming responses."""
import dataclasses
from collections import deque
from typing import Deque

from qm.grpc_client_interceptor import generic_client_interceptor
from qm.logger import logger

RECEIVED_HEADERS_MAX_SIZE = 10000


@dataclasses.dataclass
class DebugData:
    received_headers: Deque[dict] = dataclasses.field(
        default_factory=lambda: deque(maxlen=RECEIVED_HEADERS_MAX_SIZE)
    )

    def append(self, received_metadata):
        self.received_headers.append(received_metadata)


def response_metadata_collect_client_interceptor(debug_data: DebugData):
    def intercept_call(
        client_call_details, request_iterator, request_streaming, response_streaming
    ):
        def extract_response_metadata(response):
            received_metadata = {it.key: it.value for it in response.initial_metadata()}
            logger.debug("Collected response metadata: {}" % received_metadata)
            debug_data.append(received_metadata)
            return response

        return client_call_details, request_iterator, extract_response_metadata

    return generic_client_interceptor.create(intercept_call)



import json
from typing import Optional, Union, List

import numpy

from qm.QmPendingJob import QmPendingJob

from qm.QmQueue import QmQueue
from qm.QmJob import QmJob
from qm.utils import fix_object_data_type as _fix_object_data_type
from qm.logger import logger
from qm.pb.frontend_pb2 import QmDataRequest, PokeRequest, CompileRequest
from qm.pb.inc_qm_api_pb2 import HighQmApiRequest
from qm.pb.qm_manager_pb2 import CloseQuantumMachineRequest, GetRunningJobRequest
from qm.program import _Program
from qm.simulate.interface import SimulationConfig
from grpc._channel import _InactiveRpcError
from qm.utils import _level_map, _set_compiler_options
from qm._errors import CompilationException


class QuantumMachine(object):
    def __init__(self, machine_id, pb_config, config, manager):
        super(QuantumMachine, self).__init__()
        self._id = machine_id
        self.id = machine_id
        self._pb_config = pb_config
        self._config = config
        self._manager = manager
        self._frontend = manager._frontend
        self._logger = logger.getChild("job")
        self._queue = QmQueue(self, manager)

    @property
    def queue(self) -> QmQueue:
        """
        Returns the queue for the Quantum Machine
        """
        return self._queue

    @property
    def manager(self) -> "qm.QuantumMachinesManager.QuantumMachinesManager":
        """
        Returns the Quantum Machines Manager
        """
        return self._manager

    def close(self):
        """
        Closes the quantum machine.

        :return: ``True`` if the close request succeeded, raises an exception otherwise.
        """
        request = CloseQuantumMachineRequest()
        request.machineID = self._id

        result = self._frontend.CloseQuantumMachine(request)
        if not result.success:
            error_str = ""
            for err in result.errors:
                error_str = error_str + err.message + "\n"
            exception = Exception(error_str)
            exception.errors = [(it.code, it.message) for it in result.errors]
            raise exception

        return result.success

    def simulate(self, program, simulate, **kwargs):
        """
        Simulate the outputs of a deterministic QUA program.

        Equivalent to ``execute()`` with ``simulate=SimulationConfig`` (see example).

        .. note::
            A simulated job does not support calling QuantumMachine API functions.

        The following example shows a simple execution of the simulator, where the
        associated config object is omitted for brevity.

        Example::

            >>> from qm.QuantumMachinesManager import QuantumMachinesManager
            >>> from qm.qua import *
            >>> from qm.simulate import SimulationConfig
            >>>
            >>> qmManager = QuantumMachinesManager()
            >>> qm1 = qmManager.open_qm(config)
            >>>
            >>> with program() as prog:
            >>>     play('pulse1', 'element1')
            >>>
            >>> job = qm1.simulate(prog, SimulationConfig(duration=100))

        :param program: A QUA ``program()`` object to execute
        :param simulate: A ``SimulationConfig`` configuration object
        :param kwargs: additional parameteres to pass to execute
        :return: a ``QmJob`` object (see QM Job API).
        """
        kwargs["simulate"] = simulate
        try:
            job = self.execute(program, **kwargs)
        except _InactiveRpcError as e:
            raise RuntimeError("Error encountered while compiling") from e
        return job

    def execute(
        self,
        program,
        duration_limit=1000,
        data_limit=20000,
        force_execution=False,
        dry_run=False,
        **kwargs,
    ) -> QmJob:
        """
        Executes a program and returns an job object to keep track of execution and get
        results.

        .. note::

            Calling execute will halt any currently running program and clear the current
            queue. If you want to add a job to the queue, use qm.queue.add()

        :param program: A QUA ``program()`` object to execute

        :param duration_limit: This parameter is ignored as it is obsolete
        :type duration_limit: int
        :param data_limit: This parameter is ignored as it is obsolete
        :type data_limit: int
        :param force_execution: This parameter is ignored as it is obsolete
        :type force_execution: bool
        :param dry_run: This parameter is ignored as it is obsolete
        :type dry_run: bool

        :return: A ``QmJob`` object (see QM Job API).

        """
        if type(program) is not _Program:
            raise Exception("program argument must be of type qm.program.Program")

        simulation = kwargs.get("simulate", None)
        if type(simulation) is SimulationConfig:
            logger.info(
                "to simulate a program, use QuantumMachinesManager.simulate(..)"
            )
            kwargs.pop("simulate")
            return self._manager.simulate(
                self.get_config(), program, simulation, **kwargs
            )

        self._queue.clear()
        current_running_job = self.get_running_job()
        if current_running_job is not None:
            current_running_job.halt()
        pending_job = self._queue.add(program, **kwargs)
        logger.info("Executing program")
        return pending_job.wait_for_execution(timeout=5)

    def compile(self, program: _Program, **kwargs) -> str:
        """
        Compiles a QUA program to be executed later. The returned `program_id`
        can then be directly added to the queue. For a detailed explanation
        see `Precompile Jobs <https://qm-docs.qualang.io/guides/features.html#precompile-jobs>`__.

        :param program: A QUA program
        :return: a program_id str

        Example::

        >>> program_id = qm.compile(program)
        >>> pending_job = qm.queue.add_compiled(program_id)
        >>> job = pending_job.wait_for_execution()
        """
        request = CompileRequest()
        request.quantumMachineId = self.id
        ast = program.build(self._pb_config)
        request.highLevelProgram.CopyFrom(ast)

        _set_compiler_options(request, **kwargs)

        logger.info("Compiling program")

        response = self._manager._frontend.Compile(request)
        messages = [(_level_map[msg.level], msg.message) for msg in response.messages]
        program_id = response.programId

        for lvl, msg in messages:
            logger.log(lvl, msg)

        if not response.ok:
            logger.error("Compilation of program " + program_id + " failed")
            raise CompilationException(program_id)

        return program_id

    def list_controllers(self):
        """
        Gets a list with the defined controllers in this qm

        :return: The names of the controllers configured in this qm
        """
        return tuple(self.get_config()["controllers"].keys())

    def set_mixer_correction(self, mixer, intermediate_frequency, lo_frequency, values):
        """
        Sets the correction matrix for correcting gain and phase imbalances
        of an IQ mixer for the supplied intermediate frequency and LO frequency.

        :param mixer: the name of the mixer, as defined in the configuration
        :type mixer: str
        :param intermediate_frequency: the intermediate frequency for which to apply the correction matrix
        :type intermediate_frequency: int
        :param lo_frequency: the LO frequency for which to apply the correction matrix
        :type lo_frequency: int
        :param values:

            tuple is of the form (v00, v01, v10, v11) where
            the matrix is
            | v00 v01 |
            | v10 v11 |

        :type values: tuple

        .. note::

            Currently, the OPX does not support multiple mixer calibration entries.
            This function will accept IF & LO frequencies written in the config file,
            and will update the correction matrix for all of the elements with the given
            mixer/frequencies combination when the program started.

            It’s not recommended to use this method while a job is running.
            To change the calibration values for a running job,
            use job.set_element_correction
        """
        if (type(values) is not tuple and type(values) is not list) or len(values) != 4:
            raise Exception("correction values must be a tuple of 4 items")

        values = [_fix_object_data_type(x) for x in values]

        request = self._init_qm_request()

        request.setCorrection.mixer.mixer = mixer
        request.setCorrection.mixer.intermediateFrequency = abs(intermediate_frequency)
        request.setCorrection.mixer.frequencyNegative = intermediate_frequency < 0
        request.setCorrection.mixer.loFrequency = lo_frequency

        request.setCorrection.correction.v00 = values[0]
        request.setCorrection.correction.v01 = values[1]
        request.setCorrection.correction.v10 = values[2]
        request.setCorrection.correction.v11 = values[3]
        response = self._frontend.PerformQmRequest(request)
        self._handle_qm_api_response(response)

    def set_correction(self, qe, values):
        """
        **DEPRECATED** This method is deprecated. Using this method will update the correction for ALL elements that use the same mixer config
        as the supplied element. Use set_mixer_correction(..) instead.

        Sets the correction matrix for correcting gain and phase imbalances
        of an IQ mixer associated with a element.

        :param qe: the name of the element to update the correction for
        :type qe: str
        :param values:

            tuple is of the form (v00, v01, v10, v11) where
            the matrix is
            | v00 v01 |
            | v10 v11 |

        :type values: tuple
        :type values: tuple
        """
        logger.warning(
            "set_correction(..) is deprecated. use set_mixer_correction(..) instead"
        )
        if type(qe) is not str:
            raise Exception("qe must be a string")
        if (type(values) is not tuple and type(values) is not list) or len(values) != 4:
            raise Exception("correction values must be a tuple of 4 items")

        values = [_fix_object_data_type(x) for x in values]

        request = self._init_qm_request()
        request.setCorrection.qe = qe
        # request.setCorrection.correction
        request.setCorrection.correction.v00 = values[0]
        request.setCorrection.correction.v01 = values[1]
        request.setCorrection.correction.v10 = values[2]
        request.setCorrection.correction.v11 = values[3]
        response = self._frontend.PerformQmRequest(request)
        self._handle_qm_api_response(response)

    def set_frequency(self, qe, freq):
        """
        **DEPRECATED** This method is deprecated. Use set_intermediate_frequency(..) instead

         Sets the frequency of an element, at the output of the mixer, taking LO frequency into account.

        :param qe: the name of the element to update the correction for
        :param freq: the frequency to set to the given element
        :type qe: str
        :type freq: float
        """

        logger.warning(
            "set_frequency(..) is deprecated. use set_intermediate_frequency(..) instead"
        )
        if type(qe) is not str:
            raise Exception("qe must be a string")
        if not isinstance(freq, (numpy.floating, float)):
            raise Exception("freq must be a float")

        config = self.get_config()
        if qe not in config["elements"]:
            raise Exception("qe does not exist")

        element = config["elements"][qe]

        if "singleInput" in element:
            self.set_intermediate_frequency(qe, freq)
        else:
            intermediate_frequency = freq - element["mixInputs"]["lo_frequency"]
            self.set_intermediate_frequency(qe, intermediate_frequency)

    def set_intermediate_frequency(self, element, freq):
        """
        Sets the intermediate frequency of the element

        :param element: the name of the element whose intermediate frequency will be updated
        :type element: str
        :param freq: the intermediate frequency to set to the given element
        :type freq: float
        """

        logger.debug(
            "Setting element '%s' intermediate frequency to '%s'", element, freq
        )
        if type(element) is not str:
            raise TypeError("element must be a string")
        if not isinstance(freq, (numpy.floating, float)):
            raise TypeError("freq must be a float")

        freq = _fix_object_data_type(freq)

        request = self._init_qm_request()
        request.setFrequency.qe = element
        request.setFrequency.value = freq
        response = self._frontend.PerformQmRequest(request)
        self._handle_qm_api_response(response)

    def get_dc_offset_by_qe(self, qe, input):
        """
        **DEPRECATED** This method is deprecated. Use get_output_dc_offset_by_element(..) instead

        Get the current DC offset of the OPX analog output channel associated with a element.

        :param qe: the name of the element to get the correction for
        :param input: the input name as appears in the element's config
        :return: the offset, in normalized output units
        """
        logger.warning(
            "get_dc_offset_by_qe(..) is deprecated. use get_output_dc_offset_by_element(..) instead"
        )
        self.get_output_dc_offset_by_element(qe, input)

    def get_output_dc_offset_by_element(self, element, input):
        """
        Get the current DC offset of the OPX analog output channel associated with a element.

        :param element: the name of the element to get the correction for
        :param input:
            the port name as appears in the element config. Options:

            `'single'`
                for an element with a single input

            `'I'` or `'Q'`
                for an element with mixer inputs
        :return: the offset, in normalized output units
        """
        config = self.get_config()

        if element in config["elements"]:
            element_obj = config["elements"][element]
        else:
            raise Exception("Element not found")

        if "singleInput" in element_obj:
            port = element_obj["singleInput"]["port"]
        elif "mixInputs" in element_obj:
            port = element_obj["mixInputs"][input]
        else:
            raise Exception("Port not found")

        if port[0] in config["controllers"]:
            controller = config["controllers"][port[0]]
        else:
            raise Exception("Controller does not exist")

        if port[1] in controller["analog_outputs"]:
            return controller["analog_outputs"][port[1]]["offset"]
        else:
            raise Exception("Port not found")

    def set_dc_offset_by_qe(self, qe, input, offset):
        """
        ***DEPRECATED***  "This method is deprecated. Use set_output_dc_offset_by_element(..) instead." *****
        set the current DC offset of the OPX analog output channel associated with a element.

        :param qe: the name of the element to update the correction for
        :type qe: str
        :param input:
            the input name as appears in the element config. Options:

            `'single'`
                for an element with single input

            `'I'` or `'Q'`
                for an element with mixer inputs

        :type input: str
        :param offset: the dc value to set to, in normalized output units. Ranges from -0.5 to 0.5 - 2^-16 in steps of 2^-16.
        :type offset: float

        .. note::

            if the sum of the DC offset and the largest waveform data-point exceed the normalized unit range specified
            above, DAC output overflow will occur and the output will be corrupted.
        """
        logger.debug(
            "This method is deprecated. Use set_output_dc_offset_by_qe(..) instead."
        )
        return self.set_output_dc_offset_by_element(qe, input, offset)

    def set_output_dc_offset_by_element(self, element, input, offset):
        """
        Set the current DC offset of the OPX analog output channel associated with a element.

        :param element: the name of the element to update the correction for
        :type element: str
        :param input:
            the input name as appears in the element config. Options:

            `'single'`
                for an element with a single input

            `'I'` or `'Q'` or a tuple ('I', 'Q')
                for an element with mixer inputs

        :type input: Union[str, Tuple[str,str], List[str]]
        :param offset: The dc value to set to, in normalized output units. Ranges from -0.5 to 0.5 - 2^-16 in steps of 2^-16.
        :type offset: Union[float, Tuple[float,float], List[float]]

        Examples::

            >>> qm.set_output_dc_offset_by_element('flux', 'single', 0.1)
            >>> qm.set_output_dc_offset_by_element('qubit', 'I', -0.01)
            >>> qm.set_output_dc_offset_by_element('qubit', ('I', 'Q'), (-0.01, 0.05))

        .. note::

            If the sum of the DC offset and the largest waveform data-point exceed the normalized unit range specified
            above, DAC output overflow will occur and the output will be corrupted.
        """
        logger.debug(
            "Setting DC offset of input '%s' on element '%s' to '%s'",
            input,
            element,
            offset,
        )
        if type(element) is not str:
            raise TypeError("element must be a string")

        if isinstance(input, (tuple, list)):
            if not isinstance(offset, (tuple, list)) or not len(input) == len(offset):
                raise TypeError("input and offset are not of the same length")
            for i, o in zip(input, offset):
                self.set_output_dc_offset_by_element(element, i, o)
            return

        if type(input) is not str:
            raise TypeError("input must be a string or a tuple of strings")
        if offset == 0:
            offset = float(offset)
        if not isinstance(offset, (numpy.floating, float)):
            raise TypeError("offset must be a float or a tuple of floats")

        offset = _fix_object_data_type(offset)

        request = self._init_qm_request()
        request.setOutputDcOffset.qe.qe = element
        request.setOutputDcOffset.qe.port = input
        request.setOutputDcOffset.I = offset
        request.setOutputDcOffset.Q = offset
        response = self._frontend.PerformQmRequest(request)
        return self._handle_qm_api_response(response)

    def set_output_filter_by_element(
        self,
        element: str,
        input: str,
        feedforward: Union[List, numpy.ndarray, None],
        feedback: Union[List, numpy.ndarray, None],
    ):
        """
        Sets the intermediate frequency of the element

        :param element: the name of the element whose ports filters will be updated
        :param input:
            the input name as appears in the element config. Options:

            `'single'`
                for an element with single input

            `'I'` or `'Q'`
                for an element with mixer inputs

        :param feedforward: the values for the feedforward filter
        :param feedback: the values for the feedback filter
        """
        logger.debug(
            f"Setting output filter of port '{input}' on element '{element}' "
            + f"to (feedforward: {feedforward}, feedback: {feedback})"
        )
        if type(element) is not str:
            raise TypeError("element must be a string")
        if type(input) is not str:
            raise TypeError("port must be a string")
        if feedforward is not None and not isinstance(
            feedforward, (numpy.ndarray, List)
        ):
            raise TypeError("feedforward must be a list or None")

        if feedback is not None and not isinstance(feedback, (numpy.ndarray, List)):
            raise TypeError("feedback must be a list or None")

        request = self._init_qm_request()
        request.setOutputFilterTaps.qe.qe = element
        request.setOutputFilterTaps.qe.port = input
        request.setOutputFilterTaps.filter.SetInParent()
        request.setOutputFilterTaps.filter.feedforward.extend(feedforward)
        request.setOutputFilterTaps.filter.feedback.extend(feedback)
        response = self._frontend.PerformQmRequest(request)
        return self._handle_qm_api_response(response)

    def set_input_dc_offset_by_element(self, element, output, offset):
        """
        set the current DC offset of the OPX analog input channel associated with a element.

        :param element: the name of the element to update the correction for
        :type element: str
        :param output:
            the output key name as appears in the element config under 'outputs'.
        :type output: str
        :param offset: the dc value to set to, in normalized input units. Ranges from -0.5 to 0.5 - 2^-16 in steps of 2^-16.
        :type offset: float

        .. note::
            If the sum of the DC offset and the largest waveform data-point exceed the normalized unit range specified
            above, DAC output overflow will occur and the output will be corrupted.
        """
        logger.debug(
            "Setting DC offset of output '%s' on element '%s' to '%s'",
            output,
            element,
            offset,
        )
        if type(element) is not str:
            raise TypeError("element must be a string")
        if type(output) is not str:
            raise TypeError("output must be a string")
        if not isinstance(offset, (numpy.floating, float)):
            raise TypeError("offset must be a float")

        offset = _fix_object_data_type(offset)

        request = self._init_qm_request()
        request.setInputDcOffset.qe.qe = element
        request.setInputDcOffset.qe.port = output
        request.setInputDcOffset.offset = offset
        response = self._frontend.PerformQmRequest(request)
        return self._handle_qm_api_response(response)

    def get_input_dc_offset_by_element(self, element, output):
        """
        Get the current DC offset of the OPX analog input channel associated with a element.

        :param element: the name of the element to get the correction for
        :param output:
            the output key name as appears in the element config under 'outputs'.
        :return: the offset, in normalized output units
        """
        config = self.get_config()

        if element in config["elements"]:
            element_obj = config["elements"][element]
        else:
            raise Exception("Element not found")

        if "outputs" in element_obj:
            outputs = element_obj["outputs"]
        else:
            raise Exception("Element has not outputs")

        if output in outputs:
            port = outputs[output]
        else:
            raise Exception("Output does not exist")

        if port[0] in config["controllers"]:
            controller = config["controllers"][port[0]]
        else:
            raise Exception("Controller does not exist")

        if "analog_inputs" not in controller:
            raise Exception("Controller has not analog inputs defined")

        if port[1] in controller["analog_inputs"]:
            return controller["analog_inputs"][port[1]]["offset"]
        else:
            raise Exception("Port not found")

    def get_digital_delay(self, element, digital_input):
        """
        Gets the delay of the digital input of the element

        :param element: the name of the element to get the delay for
        :param digital_input: the digital input name as appears in the element's config
        :return: the delay
        """
        element_object = None
        config = self.get_config()
        for (key, value) in config["elements"].items():
            if key == element:
                element_object = value
                break

        if element_object is None:
            raise Exception("element not found")

        for (key, value) in element_object["digitalInputs"].items():
            if key == digital_input:
                return value["delay"]

        raise Exception("digital input not found")

    def set_digital_delay(self, element, digital_input, delay):
        """
        Sets the delay of the digital input of the element

        :param element: the name of the element to update delay for
        :type element: str
        :param digital_input: the digital input name as appears in the element's config
        :type digital_input: str
        :param delay: the delay value to set to, in nsec. Range: 0 to 255 - 2 * buffer, in steps of 1
        :type delay: int
        """
        logger.debug(
            "Setting delay of digital port '%s' on element '%s' to '%s'",
            digital_input,
            element,
            delay,
        )
        if type(element) is not str:
            raise Exception("element must be a string")
        if type(digital_input) is not str:
            raise Exception("port must be a string")
        if type(delay) is not int:
            raise Exception("delay must be an int")

        request = self._init_qm_request()
        request.setDigitalRoute.delay.qe = element
        request.setDigitalRoute.delay.port = digital_input
        request.setDigitalRoute.value = delay

        response = self._frontend.PerformQmRequest(request)
        return self._handle_qm_api_response(response)

    def get_digital_buffer(self, element, digital_input):
        """
        Gets the buffer for digital input of the element

        :param element: the name of the element to get the buffer for
        :type element: str
        :param digital_input: the digital input name as appears in the element's config
        :type digital_input: str
        :return: the buffer
        """
        element_object = None
        config = self.get_config()
        for (key, value) in config["elements"].items():
            if key == element:
                element_object = value
                break

        if element_object is None:
            raise Exception("element not found")

        for (key, value) in element_object["digitalInputs"].items():
            if key == digital_input:
                return value["buffer"]

        raise Exception("digital input not found")

    def set_digital_buffer(self, element, digital_input, buffer):
        """
        Sets the buffer for digital input of the element

        :param element: the name of the element to update buffer for
        :type element: str
        :param digital_input: the digital input name as appears in the element's config
        :type digital_input: str
        :param buffer: the buffer value to set to, in nsec. Range: 0 to (255 - delay) / 2, in steps of 1
        :type buffer: int
        """
        logger.debug(
            "Setting buffer of digital port '%s' on element '%s' to '%s'",
            digital_input,
            element,
            buffer,
        )
        if type(element) is not str:
            raise Exception("element must be a string")
        if type(digital_input) is not str:
            raise Exception("port must be a string")
        if type(buffer) is not int:
            raise Exception("buffer must be an int")

        request = self._init_qm_request()
        request.setDigitalRoute.buffer.qe = element
        request.setDigitalRoute.buffer.port = digital_input
        request.setDigitalRoute.value = buffer

        response = self._frontend.PerformQmRequest(request)
        return self._handle_qm_api_response(response)

    def get_time_of_flight(self, element):
        """
        Gets the *time of flight*, associated with a measurement element.

        This is the amount of time between the beginning of a measurement pulse applied to element
        and the time that the data is available to the controller for demodulation or streaming.

        :param element: the name of the element to get time of flight for
        :type element: str
        :return: the time of flight, in nsec
        """
        element_object = None
        config = self.get_config()
        for (key, value) in config["elements"].items():
            if key == element:
                if "time_of_flight" in value:
                    return value["time_of_flight"]
                else:
                    return 0

        if element_object is None:
            raise Exception("element not found")

    def get_smearing(self, element):
        """
        Gets the *smearing* associated with a measurement element.

        This is a broadening of the raw results acquisition window, to account for dispersive broadening
        in the measurement elements (readout resonators etc.) The acquisition window will be broadened
        by this amount on both sides.

        :param element: the name of the element to get smearing for
        :type element: str
        :return: the smearing, in nesc.
        """
        element_object = None
        config = self.get_config()
        for (key, value) in config["elements"].items():
            if key == element:
                if "smearing" in value:
                    return value["smearing"]
                else:
                    return 0

        if element_object is None:
            raise Exception("element not found")

    def set_io1_value(self, value_1: Union[float, bool, int]):
        """
        Sets the value of ``IO1``.

        This can be used later inside a QUA program as a QUA variable ``IO1`` without declaration.
        The type of QUA variable is inferred from the python type passed to ``value_1``, according to the following rule:

        int -> int
        float -> fixed
        bool -> bool

        :param value_1: the value to be placed in ``IO1``
        :type value_1: Union[float,bool,int]
        """
        self.set_io_values(value_1, None)

    def set_io2_value(self, value_2: Union[float, bool, int]):
        """
        Sets the value of ``IO1``.

        This can be used later inside a QUA program as a QUA variable ``IO2`` without declaration.
        The type of QUA variable is inferred from the python type passed to ``value_2``, according to the following rule:

        int -> int
        float -> fixed
        bool -> bool

        :param value_2: the value to be placed in ``IO1``
        :type value_2: Union[float, bool, int]
        """
        self.set_io_values(None, value_2)

    def set_io_values(
        self,
        value_1: Optional[Union[float, bool, int]],
        value_2: Optional[Union[float, bool, int]],
    ):
        """
        Sets the values of ``IO1`` and ``IO2``

        This can be used later inside a QUA program as a QUA variable ``IO1``, ``IO2`` without declaration.
        The type of QUA variable is inferred from the python type passed to ``value_1``, ``value_2``,
        according to the following rule:

        int -> int
        float -> fixed
        bool -> bool

        :param value_1: the value to be placed in ``IO1``
        :param value_2: the value to be placed in ``IO2``
        :type value_1: Optional[Union[float, bool, int]]
        :type value_2: Optional[Union[float, bool, int]]
        """

        if value_1 is None and value_2 is None:
            return

        if value_1 is not None:
            logger.debug("Setting value '%s' to IO1", value_1)
        if value_2 is not None:
            logger.debug("Setting value '%s' to IO2", value_2)

        request = self._init_qm_request()
        request.setIOValues.all = True

        value_1 = _fix_object_data_type(value_1)
        value_2 = _fix_object_data_type(value_2)

        if type(value_1) is int:
            request.setIOValues.ioValueSetData.add(io_number=1, intValue=value_1)
        elif type(value_1) is float:
            request.setIOValues.ioValueSetData.add(io_number=1, doubleValue=value_1)
        elif type(value_1) is bool:
            request.setIOValues.ioValueSetData.add(io_number=1, booleanValue=value_1)
        elif value_1 is None:
            pass
        else:
            raise Exception(
                "Invalid value_1 type (The possible types are: int, bool or float)"
            )

        if type(value_2) is int:
            request.setIOValues.ioValueSetData.add(io_number=2, intValue=value_2)
        elif type(value_2) is float:
            request.setIOValues.ioValueSetData.add(io_number=2, doubleValue=value_2)
        elif type(value_2) is bool:
            request.setIOValues.ioValueSetData.add(io_number=2, booleanValue=value_2)
        elif value_2 is None:
            pass
        else:
            raise Exception(
                "Invalid value_2 type (The possible types are: int, bool or float)"
            )

        response = self._frontend.PerformQmRequest(request)
        return self._handle_qm_api_response(response)

    def _init_qm_request(self):
        request = HighQmApiRequest()
        request.quantumMachineId = self._id
        return request

    @staticmethod
    def _handle_qm_api_response(response):
        if not response.ok:
            msg = "\n\t" + "\n\t".join([it.message for it in response.errors])
            logger.error("Failed: %s", msg)
            raise RuntimeError("\n".join([it.message for it in response.errors]))

    def get_io1_value(self):
        """
        Gets the data stored in ``IO1``

        No inference is made on type.

        :return:
            A dictionary with data stored in ``IO1``.
            (Data is in all three format: ``int``, ``float`` and ``bool``)
        """
        return self.get_io_values()[0]

    def get_io2_value(self):
        """
        Gets the data stored in ``IO2``

        No inference is made on type.

        :return:
            A dictionary with data from the second IO register.
            (Data is in all three format: ``int``, ``float`` and ``bool``)
        """
        return self.get_io_values()[1]

    def get_io_values(self):
        """
        Gets the data stored in both ``IO1`` and ``IO2``

        No inference is made on type.

        :return:
            A list that contains dictionaries with data from the IO registers.
            (Data is in all three format: ``int``, ``float`` and ``bool``)
        """
        request = QmDataRequest()
        request.io_value_Request.add(io_number=1, quantumMachineId=self._id)
        request.io_value_Request.add(io_number=2, quantumMachineId=self._id)
        response = self._frontend.RequestData(request)

        if not response.success:
            error_str = ""
            for err in response.errors:
                error_str = error_str + err.message + "\n"
            exception = Exception(error_str)
            exception.errors = [(it.code, it.message) for it in response.errors]
            raise exception

        resp1 = response.io_value_response[0]
        resp2 = response.io_value_response[1]

        return [
            {
                "io_number": 1,
                "int_value": resp1.values.int_value,
                "fixed_value": resp1.values.double_value,
                "boolean_value": resp1.values.boolean_value,
            },
            {
                "io_number": 2,
                "int_value": resp2.values.int_value,
                "fixed_value": resp2.values.double_value,
                "boolean_value": resp2.values.boolean_value,
            },
        ]

    def peek(self, address):
        raise NotImplementedError()
        # if you must use this, code below will work for a specific controller
        # request = PeekRequest()
        # request.controllerId = list(self._config["controllers"].keys())[0]
        # request.address = address

        # return self._frontend.Peek(request)

    def poke(self, address, value):

        request = PokeRequest()
        request.address = address
        request.value = value

        return self._frontend.Poke(request)

    def get_config(self):
        """
        Gets the current config of the qm

        :return: A dictionary with the qm's config
        """
        configs = self._manager._get_qm_config(self._id)
        self._pb_config = configs["pb_config"]
        self._config = configs["config"]
        return self._config

    def save_config_to_file(self, filename):
        """
        Saves the qm current config to a file

        :param: filename: The name of the file where the config will be saved
        """
        json_str = json.dumps(self.get_config())
        with open(filename, "w") as writer:
            writer.write(json_str)

    def get_running_job(self) -> Optional[QmJob]:
        """
        Gets the currently running job. Returns None if there isn't one.
        """
        request = GetRunningJobRequest()
        request.machineID = self._id

        response = self._frontend.GetRunningJob(request)
        if response.jobId == "":
            return None
        else:
            return QmPendingJob(response.jobId, self, self._manager).wait_for_execution(
                timeout=10
            )



import numpy as np

from qm._QmJobErrors import _handle_job_manager_error
from qm.pb.errors_pb2 import (
    MissingJobError,
    InvalidJobExecutionStatusError,
    UnknownInputStreamError,
)
from qm.pb.general_messages_pb2 import (
    Message_LEVEL_ERROR,
    Message_LEVEL_INFO,
    Message_LEVEL_WARNING,
)
from qm.logger import logger, INFO, WARN, ERROR
from qm.pb.inc_qua_pb2 import QuaProgram
from qm.pb.job_manager_pb2 import InsertInputStreamRequest


def fix_object_data_type(obj):
    if isinstance(obj, np.floating):
        return obj.item()
    else:
        return obj


_level_map = {
    Message_LEVEL_ERROR: ERROR,
    Message_LEVEL_WARNING: WARN,
    Message_LEVEL_INFO: INFO,
}


def _set_compiler_options(request, **kwargs):
    request.highLevelProgram.compilerOptions.optimizeMergeCodeExecution = (
        kwargs.get("optimize_merge_code_execution") is not False
    )
    request.highLevelProgram.compilerOptions.optimizeWriteReadCommands = (
        kwargs.get("optimize_read_write_commands") is True
    )
    request.highLevelProgram.compilerOptions.strict = (
        kwargs.get("strict", False) is True
    )

    # handle deprecated skip_optimizations (backwards compatible for future flags)
    skip_optimizations = kwargs.get("skip_optimizations", None)
    if skip_optimizations is None:
        pass
    elif type(skip_optimizations) is tuple:
        optimization_to_skip = []
        for opt in skip_optimizations:
            if type(opt) is str:
                request.highLevelProgram.compilerOptions.skipOptimizations.append(opt)
                optimization_to_skip.append(opt)
        logger.info("Skipping optimizations: " + ",".join(optimization_to_skip))
    else:
        logger.warn("skip_optimizations must be a tuple of strings")

    # handle deprecated extra_optimizations (backwards compatible for future flags)
    extra_optimizations = kwargs.get("extra_optimizations", None)
    if extra_optimizations is None:
        pass
    elif type(extra_optimizations) is tuple:
        optimization_to_add = []
        for opt in extra_optimizations:
            if type(opt) is str:
                request.highLevelProgram.compilerOptions.skipOptimizations.append(
                    "!" + opt
                )
                optimization_to_add.append(opt)
        logger.info("extra optimizations: " + ",".join(optimization_to_add))
    else:
        logger.warn("extra_optimizations must be a tuple of strings")

    try:
        flags_arg = kwargs.get("flags", [])
        flags = [opt for opt in flags_arg if type(opt) is str]
    except TypeError:
        flags = []
    for opt in flags:
        request.highLevelProgram.compilerOptions.flags.append(opt)
    logger.info("Flags: " + ",".join(flags))


def insert_input_stream_to_job(job, data, input_stream):
    if not job._qmm._caps.supports_input_stream:
        raise Exception("`insert_input_stream()` is not supported by the QOP version.")
    request = InsertInputStreamRequest()
    request.jobId = job._id
    input_stream = input_stream.unwrap()
    if isinstance(input_stream, QuaProgram.ArrayVarRefExpression):
        request.streamName = input_stream.name
    elif (
        isinstance(input_stream, QuaProgram.AnyScalarExpression)
        and input_stream.WhichOneof("expression_oneof") == "variable"
    ):
        request.streamName = input_stream.variable.name
    else:
        raise RuntimeError(
            "Unsupported type for the input stream. Only input streams declared in the QUA program can be used."
        )
    if not isinstance(data, list):
        data = [data]
    if isinstance(data[0], int):
        request.intStreamData.data.extend(data)
        request.intStreamData.SetInParent()
    elif isinstance(data[0], float):
        request.fixedStreamData.data.extend(data)
        request.fixedStreamData.SetInParent()
    elif isinstance(data[0], bool):
        request.boolStreamData.data.extend(data)
        request.boolStreamData.SetInParent()
    response = job._job_manager_client.InsertInputStream(request)
    valid_errors = (
        MissingJobError,
        InvalidJobExecutionStatusError,
        UnknownInputStreamError,
    )
    _handle_job_manager_error(request, response, valid_errors)



from qm.QuaNodeVisitor import QuaNodeVisitor


class ExpressionSerializingVisitor(QuaNodeVisitor):
    def __init__(self) -> None:
        self._out = ""
        super().__init__()

    def _default_visit(self, node):
        print("missing expression: " + node.DESCRIPTOR.full_name)
        super()._default_visit(node)

    def visit_qm_grpc_qua_QuaProgram_LibFunctionExpression(self, node):
        if node.libraryName == "random":
            args = [
                ExpressionSerializingVisitor.serialize(arg) for arg in node.arguments
            ]
            self._out = f"call_library_function('{node.libraryName}', '{node.functionName}', [{','.join(args)}])"
        else:
            library_name = {
                "util": "Util",
                "math": "Math",
                "cast": "Cast",
            }.get(node.libraryName, None)
            function_name = {
                "cond": "cond",
                "unsafe_cast_fixed": "unsafe_cast_fixed",
                "unsafe_cast_bool": "unsafe_cast_bool",
                "unsafe_cast_int": "unsafe_cast_int",
                "to_int": "to_int",
                "to_bool": "to_bool",
                "to_fixed": "to_fixed",
                "mul_fixed_by_int": "mul_fixed_by_int",
                "mul_int_by_fixed": "mul_int_by_fixed",
                "log": "log",
                "pow": "pow",
                "div": "div",
                "exp": "exp",
                "pow2": "pow2",
                "ln": "ln",
                "log2": "log2",
                "log10": "log10",
                "sqrt": "sqrt",
                "inv_sqrt": "inv_sqrt",
                "inv": "inv",
                "MSB": "msb",
                "elu": "elu",
                "aelu": "aelu",
                "selu": "selu",
                "relu": "relu",
                "plrelu": "plrelu",
                "lrelu": "lrelu",
                "sin2pi": "sin2pi",
                "cos2pi": "cos2pi",
                "abs": "abs",
                "sin": "sin",
                "cos": "cos",
                "sum": "sum",
                "max": "max",
                "min": "min",
                "argmax": "argmax",
                "argmin": "argmin",
                "dot": "dot",
            }.get(node.functionName, None)
            if library_name is None:
                raise Exception("Unsupported library name " + node.libraryName)
            if function_name is None:
                raise Exception("Unsupported function name " + node.functionName)

            args = [
                ExpressionSerializingVisitor.serialize(arg) for arg in node.arguments
            ]

            self._out = f"{library_name}.{function_name}({','.join(args)})"

    def visit_qm_grpc_qua_QuaProgram_LibFunctionExpression_Argument(self, node):
        if node.HasField("scalar"):
            self._out = ExpressionSerializingVisitor.serialize(node.scalar)
        elif node.HasField("array"):
            self._out = ExpressionSerializingVisitor.serialize(node.array)
        else:
            raise Exception("Unknown library function argument")

    def visit_qm_grpc_qua_QuaProgram_VarRefExpression(self, node):
        self._out = node.name if node.name else f"IO{node.ioNumber}"

    def visit_qm_grpc_qua_QuaProgram_ArrayVarRefExpression(self, node):
        self._out = node.name

    def visit_qm_grpc_qua_QuaProgram_ArrayCellRefExpression(self, node):
        var = ExpressionSerializingVisitor.serialize(node.arrayVar)
        index = ExpressionSerializingVisitor.serialize(node.index)
        self._out = f"{var}[{index}]"

    def visit_qm_grpc_qua_QuaProgram_LiteralExpression(self, node):
        self._out = node.value

    def visit_qm_grpc_qua_QuaProgram_AssignmentStatement_Target(self, node):
        super()._default_visit(node)

    def visit_qm_grpc_qua_QuaProgram_MeasureProcess(self, node):
        super()._default_visit(node)

    def visit_qm_grpc_qua_QuaProgram_AnalogMeasureProcess(self, node):
        super()._default_visit(node)

    def visit_qm_grpc_qua_QuaProgram_AnalogMeasureProcess_DemodIntegration(self, node):
        name = node.integration.name
        output = node.elementOutput
        target_desc, target_value = node.target.ListFields()[0]
        if target_desc.name == "scalarProcess":
            target = ExpressionSerializingVisitor.serialize(target_value)
            self._out = f'demod.full("{name}", {target}, "{output}")'
        elif target_desc.name == "vectorProcess":
            target = ExpressionSerializingVisitor.serialize(target_value.array)

            time_desc, time_value = target_value.timeDivision.ListFields()[0]
            if time_desc.name == "sliced":
                self._out = f'demod.sliced("{name}", {target}, {time_value.samplesPerChunk}, "{output}")'
            elif time_desc.name == "accumulated":
                self._out = f'demod.accumulated("{name}", {target}, {time_value.samplesPerChunk}, "{output}")'
            elif time_desc.name == "movingWindow":
                self._out = f'demod.moving_window("{name}", {target}, {time_value.samplesPerChunk}, {time_value.chunksPerWindow}, "{output}")'
            else:
                raise Exception("Unsupported analog process target " + target_desc.name)
        else:
            raise Exception("Unsupported analog process target " + target_desc.name)

    def visit_qm_grpc_qua_QuaProgram_AnalogMeasureProcess_BareIntegration(self, node):
        name = node.integration.name
        output = node.elementOutput
        target_desc, target_value = node.target.ListFields()[0]
        if target_desc.name == "scalarProcess":
            target = ExpressionSerializingVisitor.serialize(target_value)
            self._out = f'integration.full("{name}", {target}, "{output}")'
        elif target_desc.name == "vectorProcess":
            target = ExpressionSerializingVisitor.serialize(target_value.array)

            time_desc, time_value = target_value.timeDivision.ListFields()[0]
            if time_desc.name == "sliced":
                self._out = f'integration.sliced("{name}", {target}, {time_value.samplesPerChunk}, "{output}")'
            elif time_desc.name == "accumulated":
                self._out = f'integration.accumulated("{name}", {target}, {time_value.samplesPerChunk}, "{output}")'
            elif time_desc.name == "movingWindow":
                self._out = f'integration.moving_window("{name}", {target}, {time_value.samplesPerChunk}, {time_value.chunksPerWindow}, "{output}")'
            else:
                raise Exception("Unsupported analog process target " + target_desc.name)
        else:
            raise Exception("Unsupported analog process target " + target_desc.name)

    def visit_qm_grpc_qua_QuaProgram_AnalogMeasureProcess_DualDemodIntegration(
        self, node
    ):
        name1 = node.integration1.name
        name2 = node.integration2.name
        output1 = node.elementOutput1
        output2 = node.elementOutput2
        target_desc, target_value = node.target.ListFields()[0]
        if target_desc.name == "scalarProcess":
            target = ExpressionSerializingVisitor.serialize(target_value)
            self._out = f'dual_demod.full("{name1}", "{output1}", "{name2}", "{output2}", {target})'
        elif target_desc.name == "vectorProcess":
            target = ExpressionSerializingVisitor.serialize(target_value.array)

            time_desc, time_value = target_value.timeDivision.ListFields()[0]
            if time_desc.name == "sliced":
                self._out = f'dual_demod.sliced("{name1}", "{output1}", "{name2}", "{output2}", {time_value.samplesPerChunk}, {target})'
            elif time_desc.name == "accumulated":
                self._out = f'dual_demod.accumulated("{name1}", "{output1}", "{name2}", "{output2}", {time_value.samplesPerChunk}, {target})'
            elif time_desc.name == "movingWindow":
                self._out = f'dual_demod.moving_window("{name1}", "{output1}", "{name2}", "{output2}", {time_value.samplesPerChunk}, {time_value.chunksPerWindow}, {target})'
            else:
                raise Exception("Unsupported analog process target " + target_desc.name)
        else:
            raise Exception("Unsupported analog process target " + target_desc.name)

    def visit_qm_grpc_qua_QuaProgram_AnalogMeasureProcess_DualBareIntegration(
        self, node
    ):
        name1 = node.integration1.name
        name2 = node.integration2.name
        output1 = node.elementOutput1
        output2 = node.elementOutput2
        target_desc, target_value = node.target.ListFields()[0]
        if target_desc.name == "scalarProcess":
            target = ExpressionSerializingVisitor.serialize(target_value)
            self._out = f'dual_integration.full("{name1}", "{output1}", "{name2}", "{output2}", {target})'
        elif target_desc.name == "vectorProcess":
            target = ExpressionSerializingVisitor.serialize(target_value.array)

            time_desc, time_value = target_value.timeDivision.ListFields()[0]
            if time_desc.name == "sliced":
                self._out = f'dual_integration.sliced("{name1}", "{output1}", "{name2}", "{output2}", {time_value.samplesPerChunk}, {target})'
            elif time_desc.name == "accumulated":
                self._out = f'dual_integration.accumulated("{name1}", "{output1}", "{name2}", "{output2}", {time_value.samplesPerChunk}, {target})'
            elif time_desc.name == "movingWindow":
                self._out = f'dual_integration.moving_window("{name1}", "{output1}", "{name2}", "{output2}", {time_value.samplesPerChunk}, {time_value.chunksPerWindow}, {target})'
            else:
                raise Exception("Unsupported analog process target " + target_desc.name)
        else:
            raise Exception("Unsupported analog process target " + target_desc.name)

    def visit_qm_grpc_qua_QuaProgram_AnalogMeasureProcess_RawTimeTagging(self, node):
        target = ExpressionSerializingVisitor.serialize(node.target)
        target_len = ExpressionSerializingVisitor.serialize(node.targetLen)
        max_time = node.maxTime
        element_output = node.elementOutput
        self._out = f'time_tagging.analog({target}, {max_time}, {target_len}, "{element_output}")'

    def visit_qm_grpc_qua_QuaProgram_AnalogMeasureProcess_HighResTimeTagging(
        self, node
    ):
        target = ExpressionSerializingVisitor.serialize(node.target)
        target_len = ExpressionSerializingVisitor.serialize(node.targetLen)
        max_time = node.maxTime
        element_output = node.elementOutput
        self._out = f'time_tagging.high_res({target}, {max_time}, {target_len}, "{element_output}")'

    def visit_qm_grpc_qua_QuaProgram_DigitalMeasureProcess(self, node):
        super()._default_visit(node)

    def visit_qm_grpc_qua_QuaProgram_DigitalMeasureProcess_Counting(self, node):
        element_outputs = []
        for element_output in node.elementOutputs:
            element_outputs.append(f'"{element_output}"')
        element_outputs_str = ",".join(element_outputs)
        target = ExpressionSerializingVisitor.serialize(node.target)
        max_time = node.maxTime
        self._out = f"counting.digital({target}, {max_time}, ({element_outputs_str}))"

    def visit_qm_grpc_qua_QuaProgram_DigitalMeasureProcess_RawTimeTagging(self, node):
        target = ExpressionSerializingVisitor.serialize(node.target)
        target_len = ExpressionSerializingVisitor.serialize(node.targetLen)
        max_time = node.maxTime
        element_output = node.elementOutput
        self._out = f'time_tagging.digital({target}, {max_time}, {target_len}, "{element_output}")'

    def visit_qm_grpc_qua_QuaProgram_AnalogProcessTarget_ScalarProcessTarget(
        self, node
    ):
        super()._default_visit(node)

    def visit_qm_grpc_qua_QuaProgram_AnalogProcessTarget_TimeDivision(self, node):
        super()._default_visit(node)

    def visit_qm_grpc_qua_QuaProgram_AnyScalarExpression(self, node):
        super()._default_visit(node)

    def visit_qm_grpc_qua_QuaProgram_SaveStatement_Source(self, node):
        super()._default_visit(node)

    def visit_qm_grpc_qua_QuaProgram_BinaryExpression(self, node):
        left = ExpressionSerializingVisitor.serialize(node.left)
        right = ExpressionSerializingVisitor.serialize(node.right)
        sop = node.op
        if sop == node.ADD:
            op = "+"
        elif sop == node.SUB:
            op = "-"
        elif sop == node.GT:
            op = ">"
        elif sop == node.LT:
            op = "<"
        elif sop == node.LET:
            op = "<="
        elif sop == node.GET:
            op = ">="
        elif sop == node.EQ:
            op = "=="
        elif sop == node.MULT:
            op = "*"
        elif sop == node.DIV:
            op = "/"
        elif sop == node.OR:
            op = "|"
        elif sop == node.AND:
            op = "&"
        elif sop == node.XOR:
            op = "^"
        elif sop == node.SHL:
            op = "<<"
        elif sop == node.SHR:
            op = ">>"
        else:
            raise Exception("Unsupported operator " + sop)
        self._out = f"{left}{op}{right}"

    @staticmethod
    def serialize(node) -> str:
        visitor = ExpressionSerializingVisitor()
        visitor.visit(node)
        return visitor._out



from typing import Optional

from qm.ExpressionSerializingVisitor import ExpressionSerializingVisitor
from qm.QuaNodeVisitor import QuaNodeVisitor
from qm.pb.inc_qua_pb2 import QuaProgram


class QuaSerializingVisitor(QuaNodeVisitor):
    def __init__(self) -> None:
        super().__init__()
        self._indent = 0
        self._lines = []
        self.tags = []

    def out(self):
        return "\n".join(["from qm.qua import *", ""] + self._lines + [""])

    def _out_lines(self):
        return self._lines

    def _default_enter(self, node) -> bool:
        name = node.DESCRIPTOR.full_name
        statement = _statements.get(name, None)
        block = _blocks.get(name, None)
        if name not in _dont_print:
            print(f"entering {name}")
        if statement:
            self._line(statement(node))
        if block:
            self._enter_block(block(node))
        return statement is None

    def enter_qm_grpc_qua_QuaResultAnalysis(self, node) -> bool:
        if len(node.ListFields()) > 0:
            self._enter_block("with stream_processing():")
            return True
        else:
            return False

    def enter_qm_grpc_qua_QuaProgram_SaveStatement(self, node):
        if node.tag not in self.tags:
            self._line(f"{node.tag} = declare_stream()")
            self.tags.append(node.tag)
        saveLine = (
            f"save({ExpressionSerializingVisitor.serialize(node.source)}, {node.tag})"
        )
        self._line(saveLine)
        return False

    def enter_qm_grpc_qua_QuaProgram_VarDeclaration(self, node):
        args = self._get_declare_var_args(node)

        if node.isInputStream:
            self._line(
                f"{node.name} = declare_input_stream({_var_type_dec[node.type]}, {_dict_to_python_call(args)})"
            )
        else:
            self._line(
                f"{node.name} = declare({_var_type_dec[node.type]}, {_dict_to_python_call(args)})"
            )

        return False

    def _get_declare_var_args(self, node):
        size = node.size
        dim = node.dim
        args = {}
        if dim > 0:
            args["size"] = size
        if dim > 0 and len(node.value) > 0:
            args["value"] = "[" + ", ".join([_ser_exp(it) for it in node.value]) + "]"
        elif len(node.value) == 1:
            args["value"] = _ser_exp(node.value[0])
        if "value" in args and "size" in args:
            del args["size"]
        return args

    def enter_qm_grpc_qua_QuaProgram_MeasureStatement(self, node):
        if node.streamAs and node.streamAs not in self.tags:
            self._line(f"{node.streamAs} = declare_stream()")
            self.tags.append(node.streamAs)
        return self._default_enter(node)

    def visit_qm_grpc_qua_QuaProgram_ForStatement(self, node):
        if len(node.body.statements) > 0:
            super()._default_visit(node.body)
        else:
            self._line("pass")

    def visit_qm_grpc_qua_QuaProgram_ForEachStatement(self, node):
        if len(node.body.statements) > 0:
            super()._default_visit(node.body)
        else:
            self._line("pass")

    def visit_qm_grpc_qua_QuaProgram_IfStatement(self, node):
        if len(node.body.statements) > 0:
            super()._default_visit(node.body)
        else:
            self._line("pass")
        elseifs_list = node.elseifs
        for elseif in elseifs_list:
            elseif_block = elseif.body
            condition = elseif.condition
            self._leave_block()
            self._line(
                f"with elif_({ExpressionSerializingVisitor.serialize(condition)}):"
            )
            self._enter_block()
            if len(elseif_block.statements) > 0:
                super()._default_visit(elseif_block)
            else:
                self._line("pass")

        else_block = getattr(node, "else")
        if len(else_block.statements) > 0:  # Cannot ID else_() with pass
            self._leave_block()
            self._line("with else_():")
            self._enter_block()
            super()._default_visit(else_block)

    def visit_qm_grpc_qua_QuaProgram_PlayStatement(self, node):
        pulse_one_of = node.WhichOneof("pulseType")
        if pulse_one_of == "namedPulse":
            pulse = f'"{node.namedPulse.name}"'
            # node.pulse - duplicate with namedPulse
        elif pulse_one_of == "rampPulse":
            pulse = f"ramp({ExpressionSerializingVisitor.serialize(node.rampPulse)})"

        element = node.qe.name
        amp = ""
        v0 = _ser_exp(node.amp.v0)
        v1 = _ser_exp(node.amp.v1)
        v2 = _ser_exp(node.amp.v2)
        v3 = _ser_exp(node.amp.v3)
        if v0 != "":
            if v1 != "":
                amp = f"*amp({v0}, {v1}, {v2}, {v3})"
            else:
                amp = f"*amp({v0})"
        args = []
        if len(node.duration.ListFields()) == 1:
            _, value = node.duration.ListFields()[0]
            args.append(f"duration={_ser_exp(value)}")
        if len(node.condition.ListFields()) == 1:
            _, value = node.condition.ListFields()[0]
            args.append(f"condition={_ser_exp(value)}")
        if node.targetInput:
            args.append(f'target="{node.targetInput}"')
        if len(node.chirp.ListFields()) > 0:
            rate_one_of = node.chirp.WhichOneof("rate")
            if rate_one_of == "scalarRate":
                rate = ExpressionSerializingVisitor.serialize(node.chirp.scalarRate)
            elif rate_one_of == "arrayRate":
                rate = ExpressionSerializingVisitor.serialize(node.chirp.arrayRate)

            sunits = node.chirp.units
            if node.chirp.continueChirp:
                args.append(f"continue_chirp=True")
            if sunits == node.chirp.HzPerNanoSec:
                units = "Hz/nsec"
            elif sunits == node.chirp.mHzPerNanoSec:
                units = "mHz/nsec"
            elif sunits == node.chirp.uHzPerNanoSec:
                units = "uHz/nsec"
            elif sunits == node.chirp.nHzPerNanoSec:
                units = "nHz/nsec"
            elif sunits == node.chirp.pHzPerNanoSec:
                units = "pHz/nsec"
            else:
                raise Exception("Unsupported units " + sunits)

            timesBuilder = []
            for time in node.chirp.times:
                timesBuilder.append(f"{time}")

            if len(timesBuilder) > 0:
                times = f'[{", ".join(timesBuilder)}]'
            else:
                times = "None"

            args.append(f'chirp=({rate},{times},"{units}")')
        if len(node.truncate.ListFields()) == 1:
            _, value = node.truncate.ListFields()[0]
            args.append(f"truncate={_ser_exp(value)}")

        # TODO maybe make sure no other fields?

        if len(args) > 0:
            args_str = f', {", ".join(args)}'
        else:
            args_str = ""
        self._line(f'play({pulse}{amp}, "{element}"{args_str})')

    def _default_leave(self, node):
        name = node.DESCRIPTOR.full_name
        block = _blocks.get(name, None)
        if block is not None:
            self._leave_block()
        super()._default_leave(node)

    def _enter_block(self, line: Optional[str] = None):
        if line is not None:
            self._line(line)
        self._indent += 1

    def _leave_block(self):
        self._indent -= 1

    def _line(self, line: str):
        self._lines.append((self._indent * "    ") + line)


def _ser_exp(value):
    return ExpressionSerializingVisitor.serialize(value)


def _dict_to_python_call(d: dict) -> str:
    return ", ".join([f"{k}={v}" for k, v in d.items()])


def _ramp_to_zero_statement(node):
    args = []
    args.append(f'"{node.qe.name}"')
    if len(node.duration.ListFields()) == 1:
        args.append(f"{node.duration.value}")
    return f'ramp_to_zero({", ".join(args)})'


def _measure_statement(node):
    args = []

    amp = ""
    v0 = _ser_exp(node.amp.v0)
    v1 = _ser_exp(node.amp.v1)
    v2 = _ser_exp(node.amp.v2)
    v3 = _ser_exp(node.amp.v3)
    if v0 != "":
        if v1 != "":
            amp = f"*amp({v0}, {v1}, {v2}, {v3})"
        else:
            amp = f"*amp({v0})"

    args.append(f'"{node.pulse.name}"{amp}')
    args.append(f'"{node.qe.name}"')
    if node.streamAs:
        args.append(f"{node.streamAs}")
    else:
        args.append(f"None")
    if len(node.measureProcesses) > 0:
        for process in node.measureProcesses:
            args.append(ExpressionSerializingVisitor.serialize(process))
    return f'measure({", ".join(args)})'


def _wait_statement(node):
    args = []
    if len(node.time.ListFields()) == 1:
        _, value = node.time.ListFields()[0]
        args.append(f"{ExpressionSerializingVisitor.serialize(value)}")
    qes = []
    for qe in node.qe:
        qes.append(f'"{qe.name}"')
    args.append(", ".join(qes))
    return f'wait({", ".join(args)})'


def _align_statement(node):
    args = []
    for qe in node.qe:
        args.append(f'"{qe.name}"')
    return f'align({", ".join(args)})'


def _wait_for_trigger_statement(node):
    args = []
    for qe in node.qe:
        args.append(f'"{qe.name}"')
    if node.pulseToPlay.name:
        args.append(f'"{node.pulseToPlay.name}"')
    if node.elementOutput.element:
        if node.elementOutput.output:
            args.append(
                f'trigger_element=("{node.elementOutput.element}", "{node.elementOutput.output}")'
            )
        else:
            args.append(f'trigger_element="{node.elementOutput.element}"')
    return f'wait_for_trigger({", ".join(args)})'


def _frame_rotation_statement(node):
    args = []
    args.append(f"{ExpressionSerializingVisitor.serialize(node.value)}")
    args.append(f'"{node.qe.name}"')
    return f'frame_rotation_2pi({", ".join(args)})'


def _reset_frame_statement(node):
    args = []
    args.append(f'"{node.qe.name}"')
    return f'reset_frame({", ".join(args)})'


def _update_frequency_statement(node):
    args = []
    args.append(f'"{node.qe.name}"')
    args.append(f"{ExpressionSerializingVisitor.serialize(node.value)}")
    if node.units == QuaProgram.UpdateFrequencyStatement.Hz:
        args.append(f'"Hz"')
    elif node.units == QuaProgram.UpdateFrequencyStatement.mHz:
        args.append(f'"mHz"')
    elif node.units == QuaProgram.UpdateFrequencyStatement.uHz:
        args.append(f'"uHz"')
    elif node.units == QuaProgram.UpdateFrequencyStatement.nHz:
        args.append(f'"nHz"')
    elif node.units == QuaProgram.UpdateFrequencyStatement.pHz:
        args.append(f'"pHz"')
    else:
        raise RuntimeError(f'unknown units "{node.units}"')
    args.append(f"{node.keepPhase}")
    return f'update_frequency({", ".join(args)})'


def _set_dc_offset_statement(node):
    args = []
    args.append(f'"{node.qe.name}"')
    args.append(f'"{node.qeInputReference}"')
    args.append(f"{ExpressionSerializingVisitor.serialize(node.offset)}")
    return f'set_dc_offset({", ".join(args)})'


def _advance_input_stream_statement(node):
    if node.streamArray.name != "":
        input_stream = ExpressionSerializingVisitor.serialize(node.streamArray)
    elif node.streamVariable.name != "":
        input_stream = ExpressionSerializingVisitor.serialize(node.streamVariable)
    else:
        raise RuntimeError("unsupported type for pop input stream")
    return f"advance_input_stream({input_stream})"


def _for_block_statement(node):
    condition = ExpressionSerializingVisitor.serialize(node.condition)
    if len(node.init.statements) == 0 and len(node.update.statements) == 0:
        if condition == "True":
            return f"with infinite_loop_():"
        else:
            return f"with while_({condition}):"
    else:
        if len(node.init.statements) != 1:
            raise Exception("for is not valid")
        if len(node.update.statements) != 1:
            raise Exception("for is not valid")

        init = node.init.statements[0].assign
        update = node.update.statements[0].assign

        if init is None:
            raise Exception("for is not valid")
        if update is None:
            raise Exception("for is not valid")

        return f"with for_({ExpressionSerializingVisitor.serialize(init.target)},{ExpressionSerializingVisitor.serialize(init.expression)},{condition},{ExpressionSerializingVisitor.serialize(update.expression)}):"


def _for_each_block_statement(node):
    variables = []
    arrays = []
    for it in node.iterator:
        variables.append(_ser_exp(it.variable))
        arrays.append(it.array.name)
    return f'with for_each_(({",".join(variables)}),({",".join(arrays)})):'


def _if_block_statement(node):
    condition = ExpressionSerializingVisitor.serialize(node.condition)
    if node.unsafe is True:
        unsafe = ", unsafe=True"
    else:
        unsafe = ""
    return f"with if_({condition}{unsafe}):"


def _stream_processing_function(array):
    function = array[0].string_value

    if function == "average":
        return f"average()"

    if function == "dot":
        if len(array) == 1:
            return f"tuple_dot_product()"
        else:
            vector = _stream_processing_operator(array[1].list_value.values)
            return f"dot_product({vector})"

    if function == "vmult":
        vector = _stream_processing_operator(array[1].list_value.values)
        return f"multiply_by({vector})"

    if function == "smult":
        return f"multiply_by({array[1].string_value})"

    if function == "tmult":
        return f"tuple_multiply()"

    if function == "conv":
        if len(array) == 2:
            if array[1].string_value:
                mode = f'"{array[1].string_value}"'
            else:
                mode = ""
            return f"tuple_convolution({mode})"
        else:
            vector = _stream_processing_operator(array[2].list_value.values)
            if array[1].string_value:
                mode = f',"{array[1].string_value}"'
            else:
                mode = ""
            return f"convolution({vector}{mode})"

    if function == "fft":
        return f"fft()"

    if function == "booleancast":
        return f"boolean_to_int()"

    if function == "demod":
        if len(array[2].list_value) > 0:
            cos = _stream_processing_operator(array[2].list_value.values)
        else:
            cos = array[2].string_value
        if len(array[3].list_value) > 0:
            sin = _stream_processing_operator(array[3].list_value.values)
        else:
            sin = array[3].string_value
        return f"demod({array[1].string_value},{cos},{sin})"

    print(f"missing function: {function}")
    return f"default_function()"


def _stream_processing_operator(array):
    operator = array[0].string_value

    if operator == "@macro_input":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.input{array[1].string_value}()"

    if operator == "@array":
        values = []
        for a in array[1:]:
            if len(a.list_value) > 0:
                value = _stream_processing_operator(a.list_value.values)
            else:
                value = a.string_value
            values.append(value)
        return f'[{", ".join(values)}]'

    if operator == "@macro_auto_reshape":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.auto_reshape()"

    if operator == "+":
        left = _stream_processing_statement(array[1])
        right = _stream_processing_statement(array[2])
        return f"{left}.add({right})"  # arithmetic stream

    if operator == "-":
        left = _stream_processing_statement(array[1])
        right = _stream_processing_statement(array[2])
        return f"{left}.subtract({right})"  # arithmetic stream

    if operator == "/":
        left = _stream_processing_statement(array[1])
        right = _stream_processing_statement(array[2])
        return f"{left}.divide({right})"  # arithmetic stream

    if operator == "*":
        left = _stream_processing_statement(array[1])
        right = _stream_processing_statement(array[2])
        return f"{left}.multiply({right})"  # arithmetic stream

    if operator == "zip":
        left = _stream_processing_statement(array[1])
        right = _stream_processing_statement(array[2])
        return f"{right}.zip({left})"  # arithmetic stream

    if operator == "take":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.take({array[1].string_value})"

    if operator == "buffer":
        chain = _default_stream_processing_chain(array)
        dims = [dim.string_value for dim in array[1:-1]]
        return f'{chain}.buffer({", ".join(dims)})'

    if operator == "bufferAndSkip":
        chain = _default_stream_processing_chain(array)
        return (
            f"{chain}.buffer_and_skip({array[1].string_value}, {array[2].string_value})"
        )

    if operator == "skip":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.skip({array[1].string_value})"

    if operator == "skipLast":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.skip_last({array[1].string_value})"

    if operator == "histogram":
        chain = _default_stream_processing_chain(array)
        bins = _stream_processing_operator(array[1].list_value.values)
        return f"{chain}.histogram({bins})"

    if operator == "average":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.average()"

    if operator == "flatten":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.flatten()"

    if operator == "map":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.map(FUNCTIONS.{_stream_processing_function(array[1].list_value.values)})"

    if operator == "@re":
        chain = _default_stream_processing_chain(array)
        timestamp_mode = int(array[1].string_value)
        if timestamp_mode == 0:  # values
            return f"{chain}"
        elif timestamp_mode == 1:  # timestamps
            return f"{chain}.timestamps()"
        elif timestamp_mode == 2:  # ValuesAndTimestamps
            return f"{chain}.with_timestamps()"

    print(f"missing operator: {operator}")
    chain = _default_stream_processing_chain(array)
    return f"{chain}"


def _default_stream_processing_chain(array):
    lastIndex = len(array) - 1
    chain = _stream_processing_statement(array[lastIndex])
    return f"{chain}"


def _stream_processing_statement(node):
    if len(node.list_value) > 0:
        return _stream_processing_operator(node.list_value.values)
    else:
        return node.string_value


def _stream_processing_terminal_statement(node):
    lastIndex = len(node.values) - 1
    chain = _stream_processing_statement(node.values[lastIndex])
    terminal = node.values[0].string_value
    terminal = "save_all" if terminal == "saveAll" else terminal  # normalize save all
    return f'{chain}.{terminal}("{node.values[1].string_value}")'


def _serialize(node):
    visitor = QuaSerializingVisitor()
    visitor.visit(node)
    return visitor._out_lines()


_blocks = {
    "qm.grpc.qua.QuaProgram": lambda n: "with program() as prog:",
    "qm.grpc.qua.QuaProgram.ForStatement": _for_block_statement,
    "qm.grpc.qua.QuaProgram.ForEachStatement": _for_each_block_statement,
    "qm.grpc.qua.QuaProgram.IfStatement": _if_block_statement,
}

_statements = {
    "qm.grpc.qua.QuaProgram.MeasureStatement": _measure_statement,
    "qm.grpc.qua.QuaProgram.WaitStatement": _wait_statement,
    "qm.grpc.qua.QuaProgram.AssignmentStatement": lambda n: f"assign({ExpressionSerializingVisitor.serialize(n.target)}, "
    f"{ExpressionSerializingVisitor.serialize(n.expression)})",
    "qm.grpc.qua.QuaProgram.PauseStatement": lambda n: "pause()",
    "qm.grpc.qua.QuaProgram.ResetPhaseStatement": lambda n: f'reset_phase("{n.qe.name}")',
    "qm.grpc.qua.QuaProgram.UpdateFrequencyStatement": _update_frequency_statement,
    "qm.grpc.qua.QuaProgram.AlignStatement": _align_statement,
    "qm.grpc.qua.QuaProgram.WaitForTriggerStatement": _wait_for_trigger_statement,
    "qm.grpc.qua.QuaProgram.ZRotationStatement": _frame_rotation_statement,
    "qm.grpc.qua.QuaProgram.RampToZeroStatement": _ramp_to_zero_statement,
    "qm.grpc.qua.QuaProgram.ResetFrameStatement": _reset_frame_statement,
    "google.protobuf.ListValue": _stream_processing_terminal_statement,
    # list value statement is assumed just as stream processing for now
    "qm.grpc.qua.QuaProgram.UpdateCorrectionStatement": lambda n: f'update_correction("{n.qe.name}"'
    f",{ExpressionSerializingVisitor.serialize(n.correction.c0)}"
    f",{ExpressionSerializingVisitor.serialize(n.correction.c1)}"
    f",{ExpressionSerializingVisitor.serialize(n.correction.c2)}"
    f",{ExpressionSerializingVisitor.serialize(n.correction.c3)})",
    "qm.grpc.qua.QuaProgram.SetDcOffsetStatement": _set_dc_offset_statement,
    "qm.grpc.qua.QuaProgram.AdvanceInputStreamStatement": _advance_input_stream_statement,
}

_var_type_dec = {0: "int", 1: "bool", 2: "fixed"}

_nodes_to_ignore = [
    "qm.grpc.qua.QuaProgram.Script",
    "qm.grpc.qua.QuaProgram.StatementsCollection",
    "qm.grpc.qua.QuaProgram.AnyStatement",
    "qm.grpc.qua.QuaProgram.AnyScalarExpression",
    "qm.grpc.qua.QuaProgram.BinaryExpression",
    "qm.grpc.qua.QuaProgram.VarRefExpression",
    "qm.grpc.qua.QuaProgram.LiteralExpression",
    "qm.grpc.qua.QuaProgram.PlayStatement",
    "qm.grpc.qua_config.QuaConfig",
]

_dont_print = set(list(_blocks.keys()) + list(_statements.keys()) + _nodes_to_ignore)



from marshmallow_polyfield import PolyField

from qm.pb.frontend_pb2 import ExecutionOverrides
from marshmallow import Schema, fields, post_load, ValidationError


def load_overrides(overrides):
    return ExecutionOverridesSchema().load(overrides)


def _non_empty(v):
    if len(v) == 0:
        raise ValidationError("List must be non-empty")


def _waveforms_schema_deserialization_disambiguation(data, _o):
    if isinstance(data, list):
        return fields.List(fields.Float(), required=True, validate=_non_empty)
    else:
        return fields.Float(required=True)


class ExecutionOverridesSchema(Schema):
    waveforms = fields.Dict(
        keys=fields.String(),
        values=PolyField(
            deserialization_schema_selector=_waveforms_schema_deserialization_disambiguation,
            required=True,
        ),
    )

    @post_load(pass_many=False)
    def build(self, data, **kwargs):
        result = ExecutionOverrides()

        try:
            waveforms = data["waveforms"]
        except KeyError:
            waveforms = {}

        for k, v in waveforms.items():
            iw = result.waveforms.get_or_create(k)
            samples = v if isinstance(v, list) else [v]
            for s in samples:
                iw.samples.append(s)

        return result



from dataclasses import dataclass
from typing import List, Dict
import numpy as np


@dataclass
class IterationData:
    iteration_variable_name: str
    iteration_values: np.ndarray


@dataclass
class StreamMetadataError:
    error: str
    location: str


@dataclass
class StreamMetadata:
    stream_name: str
    iteration_values: List[IterationData]


def _get_numpy_array_from_proto_iteration_data(iteration_data) -> np.ndarray:
    if iteration_data.HasField("forEachIntIterationValues"):
        return np.array(iteration_data.forEachIntIterationValues.values)
    elif iteration_data.HasField("forEachDoubleIterationValues"):
        return np.array(iteration_data.forEachDoubleIterationValues.values)
    elif iteration_data.HasField("forDoubleIterationValues"):
        start = iteration_data.forDoubleIterationValues.startValue
        step = iteration_data.forDoubleIterationValues.step
        stop = start + step * iteration_data.forDoubleIterationValues.numberOfIterations
        return np.arange(start=start, step=step, stop=round(stop, 9))
    elif iteration_data.HasField("forIntIterationValues"):
        start = iteration_data.forIntIterationValues.startValue
        step = iteration_data.forIntIterationValues.step
        stop = (
            iteration_data.forIntIterationValues.startValue
            + iteration_data.forIntIterationValues.step
            * iteration_data.forIntIterationValues.numberOfIterations
        )
        return np.arange(start=start, step=step, stop=stop)


def _get_stream_metadata_dict_from_proto_resp(
    program_metadata,
) -> Dict[str, List[StreamMetadata]]:
    stream_metadata_dict = {
        x.streamName: StreamMetadata(
            x.streamName,
            [
                IterationData(
                    y.iterationVariableName,
                    _get_numpy_array_from_proto_iteration_data(y),
                )
                for y in x.iterationData
            ],
        )
        for x in program_metadata.programStreamMetadata.streamMetadata
    }

    return stream_metadata_dict



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: qm/pb/compiler.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from qm.pb import inc_qua_pb2 as qm_dot_pb_dot_inc__qua__pb2
from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from qm.pb import qm_manager_pb2 as qm_dot_pb_dot_qm__manager__pb2
from qm.pb import general_messages_pb2 as qm_dot_pb_dot_general__messages__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='qm/pb/compiler.proto',
  package='qm.grpc.compiler',
  syntax='proto3',
  serialized_options=b'\n\020qm.grpc.compilerB\rCompilerProtoP\001\242\002\003HLW',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x14qm/pb/compiler.proto\x12\x10qm.grpc.compiler\x1a\x13qm/pb/inc_qua.proto\x1a\x1cgoogle/protobuf/struct.proto\x1a\x16qm/pb/qm_manager.proto\x1a\x1cqm/pb/general_messages.proto\"K\n\tQuaValues\x12\x11\n\tint_value\x18\x01 \x01(\r\x12\x14\n\x0c\x64ouble_value\x18\x02 \x01(\x01\x12\x15\n\rboolean_value\x18\x03 \x01(\x08\"I\n\x0e\x43ompileRequest\x12(\n\x07program\x18\x01 \x01(\x0b\x32\x17.qm.grpc.qua.QuaProgram\x12\r\n\x05jobId\x18\x02 \x01(\t\"\xdf\x01\n\x0f\x43ompileResponse\x12\n\n\x02ok\x18\x01 \x01(\x08\x12\x33\n\x08messages\x18\x02 \x03(\x0b\x32!.qm.grpc.compiler.CompilerMessage\x12\x0e\n\x06\x62inary\x18\x03 \x01(\x0c\x12\x10\n\x08metadata\x18\x04 \x01(\t\x12;\n\x05\x64\x65\x62ug\x18\x1e \x03(\x0b\x32,.qm.grpc.compiler.CompileResponse.DebugEntry\x1a,\n\nDebugEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"Y\n\x0f\x43ompilerMessage\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\x35\n\x05level\x18\x02 \x01(\x0e\x32&.qm.grpc.general_messages.MessageLevel\"S\n\x12ValidationResponse\x12=\n\x08messages\x18\x01 \x03(\x0b\x32+.qm.grpc.qm_manager.ConfigValidationMessage\"G\n\rDynamicConfig\x12\x0f\n\x07version\x18\x01 \x01(\r\x12%\n\x04root\x18\x02 \x01(\x0b\x32\x17.google.protobuf.StructB)\n\x10qm.grpc.compilerB\rCompilerProtoP\x01\xa2\x02\x03HLWb\x06proto3'
  ,
  dependencies=[qm_dot_pb_dot_inc__qua__pb2.DESCRIPTOR,google_dot_protobuf_dot_struct__pb2.DESCRIPTOR,qm_dot_pb_dot_qm__manager__pb2.DESCRIPTOR,qm_dot_pb_dot_general__messages__pb2.DESCRIPTOR,])




_QUAVALUES = _descriptor.Descriptor(
  name='QuaValues',
  full_name='qm.grpc.compiler.QuaValues',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='int_value', full_name='qm.grpc.compiler.QuaValues.int_value', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='double_value', full_name='qm.grpc.compiler.QuaValues.double_value', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='boolean_value', full_name='qm.grpc.compiler.QuaValues.boolean_value', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=147,
  serialized_end=222,
)


_COMPILEREQUEST = _descriptor.Descriptor(
  name='CompileRequest',
  full_name='qm.grpc.compiler.CompileRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='program', full_name='qm.grpc.compiler.CompileRequest.program', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='jobId', full_name='qm.grpc.compiler.CompileRequest.jobId', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=224,
  serialized_end=297,
)


_COMPILERESPONSE_DEBUGENTRY = _descriptor.Descriptor(
  name='DebugEntry',
  full_name='qm.grpc.compiler.CompileResponse.DebugEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='qm.grpc.compiler.CompileResponse.DebugEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='qm.grpc.compiler.CompileResponse.DebugEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=479,
  serialized_end=523,
)

_COMPILERESPONSE = _descriptor.Descriptor(
  name='CompileResponse',
  full_name='qm.grpc.compiler.CompileResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='ok', full_name='qm.grpc.compiler.CompileResponse.ok', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='messages', full_name='qm.grpc.compiler.CompileResponse.messages', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='binary', full_name='qm.grpc.compiler.CompileResponse.binary', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='metadata', full_name='qm.grpc.compiler.CompileResponse.metadata', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='debug', full_name='qm.grpc.compiler.CompileResponse.debug', index=4,
      number=30, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_COMPILERESPONSE_DEBUGENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=300,
  serialized_end=523,
)


_COMPILERMESSAGE = _descriptor.Descriptor(
  name='CompilerMessage',
  full_name='qm.grpc.compiler.CompilerMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='message', full_name='qm.grpc.compiler.CompilerMessage.message', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='level', full_name='qm.grpc.compiler.CompilerMessage.level', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=525,
  serialized_end=614,
)


_VALIDATIONRESPONSE = _descriptor.Descriptor(
  name='ValidationResponse',
  full_name='qm.grpc.compiler.ValidationResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='messages', full_name='qm.grpc.compiler.ValidationResponse.messages', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=616,
  serialized_end=699,
)


_DYNAMICCONFIG = _descriptor.Descriptor(
  name='DynamicConfig',
  full_name='qm.grpc.compiler.DynamicConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='version', full_name='qm.grpc.compiler.DynamicConfig.version', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='root', full_name='qm.grpc.compiler.DynamicConfig.root', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=701,
  serialized_end=772,
)

_COMPILEREQUEST.fields_by_name['program'].message_type = qm_dot_pb_dot_inc__qua__pb2._QUAPROGRAM
_COMPILERESPONSE_DEBUGENTRY.containing_type = _COMPILERESPONSE
_COMPILERESPONSE.fields_by_name['messages'].message_type = _COMPILERMESSAGE
_COMPILERESPONSE.fields_by_name['debug'].message_type = _COMPILERESPONSE_DEBUGENTRY
_COMPILERMESSAGE.fields_by_name['level'].enum_type = qm_dot_pb_dot_general__messages__pb2._MESSAGELEVEL
_VALIDATIONRESPONSE.fields_by_name['messages'].message_type = qm_dot_pb_dot_qm__manager__pb2._CONFIGVALIDATIONMESSAGE
_DYNAMICCONFIG.fields_by_name['root'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
DESCRIPTOR.message_types_by_name['QuaValues'] = _QUAVALUES
DESCRIPTOR.message_types_by_name['CompileRequest'] = _COMPILEREQUEST
DESCRIPTOR.message_types_by_name['CompileResponse'] = _COMPILERESPONSE
DESCRIPTOR.message_types_by_name['CompilerMessage'] = _COMPILERMESSAGE
DESCRIPTOR.message_types_by_name['ValidationResponse'] = _VALIDATIONRESPONSE
DESCRIPTOR.message_types_by_name['DynamicConfig'] = _DYNAMICCONFIG
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

QuaValues = _reflection.GeneratedProtocolMessageType('QuaValues', (_message.Message,), {
  'DESCRIPTOR' : _QUAVALUES,
  '__module__' : 'qm.pb.compiler_pb2'
  # @@protoc_insertion_point(class_scope:qm.grpc.compiler.QuaValues)
  })
_sym_db.RegisterMessage(QuaValues)

CompileRequest = _reflection.GeneratedProtocolMessageType('CompileRequest', (_message.Message,), {
  'DESCRIPTOR' : _COMPILEREQUEST,
  '__module__' : 'qm.pb.compiler_pb2'
  # @@protoc_insertion_point(class_scope:qm.grpc.compiler.CompileRequest)
  })
_sym_db.RegisterMessage(CompileRequest)

CompileResponse = _reflection.GeneratedProtocolMessageType('CompileResponse', (_message.Message,), {

  'DebugEntry' : _reflection.GeneratedProtocolMessageType('DebugEntry', (_message.Message,), {
    'DESCRIPTOR' : _COMPILERESPONSE_DEBUGENTRY,
    '__module__' : 'qm.pb.compiler_pb2'
    # @@protoc_insertion_point(class_scope:qm.grpc.compiler.CompileResponse.DebugEntry)
    })
  ,
  'DESCRIPTOR' : _COMPILERESPONSE,
  '__module__' : 'qm.pb.compiler_pb2'
  # @@protoc_insertion_point(class_scope:qm.grpc.compiler.CompileResponse)
  })
_sym_db.RegisterMessage(CompileResponse)
_sym_db.RegisterMessage(CompileResponse.DebugEntry)

CompilerMessage = _reflection.GeneratedProtocolMessageType('CompilerMessage', (_message.Message,), {
  'DESCRIPTOR' : _COMPILERMESSAGE,
  '__module__' : 'qm.pb.compiler_pb2'
  # @@protoc_insertion_point(class_scope:qm.grpc.compiler.CompilerMessage)
  })
_sym_db.RegisterMessage(CompilerMessage)

ValidationResponse = _reflection.GeneratedProtocolMessageType('ValidationResponse', (_message.Message,), {
  'DESCRIPTOR' : _VALIDATIONRESPONSE,
  '__module__' : 'qm.pb.compiler_pb2'
  # @@protoc_insertion_point(class_scope:qm.grpc.compiler.ValidationResponse)
  })
_sym_db.RegisterMessage(ValidationResponse)

DynamicConfig = _reflection.GeneratedProtocolMessageType('DynamicConfig', (_message.Message,), {
  'DESCRIPTOR' : _DYNAMICCONFIG,
  '__module__' : 'qm.pb.compiler_pb2'
  # @@protoc_insertion_point(class_scope:qm.grpc.compiler.DynamicConfig)
  })
_sym_db.RegisterMessage(DynamicConfig)


DESCRIPTOR._options = None
_COMPILERESPONSE_DEBUGENTRY._options = None
# @@protoc_insertion_point(module_scope)



from typing.io import BinaryIO
from pathlib import Path


class BinaryAsset:
    def for_writing(self) -> BinaryIO:
        raise NotImplementedError()

    def for_reading(self) -> BinaryIO:
        raise NotImplementedError()


class BaseStore:
    """
    The interface to saving data from a running QM job
    """

    def __init__(self) -> None:
        super().__init__()

    def job_named_result(self, job_id: str, name: str) -> BinaryAsset:
        raise NotImplementedError()

    def all_job_results(self, job_id: str) -> BinaryAsset:
        raise NotImplementedError()


class FileBinaryAsset(BinaryAsset):
    def __init__(self, path: Path) -> None:
        super().__init__()
        self._path = path

    def for_writing(self) -> BinaryIO:
        return self._path.open("wb")

    def for_reading(self) -> BinaryIO:
        return self._path.open("rb")


class SimpleFileStore(BaseStore):
    def __init__(self, root: str = ".") -> None:
        super().__init__()
        self._root = Path(root).absolute()

    def _job_path(self, job_id: str):
        path = Path(f"{self._root}/{job_id}")
        path.mkdir(parents=True, exist_ok=True)
        return path

    def job_named_result(self, job_id: str, name: str) -> BinaryAsset:
        return FileBinaryAsset(self._job_path(job_id).joinpath(f"result_{name}.npy"))

    def all_job_results(self, job_id: str) -> BinaryAsset:
        return FileBinaryAsset(self._job_path(job_id).joinpath(f"results.npz"))
